/*    Filename: periodic_cpu_load_single.c
 *      Author: Darren Hart <dvhltc@us.ibm.com>
 * Description: Measure variation in computational execution time
 *              at the specified period, priority, and loops.
 *
 * Use "-j" to enable the jvm simulator.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (C) IBM Corporation, 2006
 *
 * 2007-May-2:		Initial version by Darren Hart <dvhltc@us.ibm.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "librt.h"
#include "libstats.h"
#include "libjvmsim.h"

//#define ITERATIONS 100 /* short functional test run */
#define ITERATIONS 10000
//#define ITERATIONS 1000 /* min iters for 3 nines */
// FIXME: need some kind of passing criteria calculation
//#define PASS_US 100
// FIXME: consider removing the histograms entirely
#define HIST_BUCKETS 100

// define sane defaults
# define DEFAULT_PERIOD 5
# define DEFAULT_PRIO   90
# define DEFAULT_CALC_LOOPS  1000000

static int prio;
static int period;
static int calc_loops;
static int run_jvmsim = 0;
static int ret = 0;

struct periodic_arg {
	int period;
	int iterations;
	void*(*func)(void*);
	void *arg;
};

void usage()
{
	rt_help();
	printf("periodic_cpu_load specific options:\n");
	printf("  -j(0,1)	0:disable jvmsim, 1:enable jvmsim\n");
	printf("  -r[0-99]	real-time priority\n");
	printf("  -tPERIOD	period in ms\n");
	printf("  -cCALC_LOOPS	loops per iteration\n");
}

void *calc(void *arg)
{
	int i, j;
	int loops = (int)arg;
	for (i = 0; i < loops; i++) {
		for (j = 0; j < 125; j++ ) {
			// Sum of the numbers up to J
			int temp = j * ( j + 1 ) / 2;
		}
	}
	return NULL;
}

void *periodic_thread(void *thread)
{
	struct thread *t = (struct thread *)thread;
	struct periodic_arg *parg = (struct periodic_arg *)t->arg;
	nsec_t period = parg->period;
	void*(*func)(void*) = parg->func;

	stats_container_t dat;
	stats_container_t hist;
	stats_quantiles_t quantiles;

	int i = 0;
	int fail = 0;
	nsec_t next;
	nsec_t exe_start, exe_end, exe_time;

	stats_container_init(&dat, ITERATIONS);
	stats_container_init(&hist, HIST_BUCKETS);
	stats_quantiles_init(&quantiles, 4);

	next = rt_gettime();
	while (i < parg->iterations) {
		next += period;
		if (rt_gettime() > next) {
			printf("TID %d missed period, aborting\n", t->id);
			fail = 1;
			break;
		}
		exe_start = rt_gettime();
		func(parg->arg);
		exe_end = rt_gettime();
		exe_time = exe_end - exe_start;
		dat.records[i].x = i;
		dat.records[i].y = exe_time/NS_PER_US;

		i++;
		rt_nanosleep(next - rt_gettime());
	}
	stats_quantiles_calc(&dat, &quantiles);

	printf("\nTID %d (prio %d) complete\n", t->id, t->priority);
	printf("  Execution Time Statistics:\n");
	printf("        Min: %ld us\n", stats_min(&dat));
	printf("        Max: %ld us\n", stats_max(&dat));
	printf("        Avg: %f us\n", stats_avg(&dat));
	printf("     StdDev: %f us\n\n", stats_stddev(&dat));
	printf("  Quantiles:\n");
	stats_quantiles_print(&quantiles);
	printf("Criteria: no missed periods\n");
	printf("Result: %s\n", fail ? "FAIL":"PASS");

	return NULL;
}

int parse_args(int c, char *v)
{
	int handled = 1;
	switch (c) {
		case 'j':
			run_jvmsim = 1;
			break;
		case 'h':
			usage();
			exit(0);
		case 'r':
			prio = atoi(v);
			break;
		case 't':
			period = atoi(v) * NS_PER_MS;
			break;
		case 'c':
			calc_loops = atoi(v);
			break;
		default:
			handled = 0;
			break;
	}
	return handled;
}

int main(int argc, char *argv[])
{
	period = DEFAULT_PERIOD*NS_PER_MS;
	prio = DEFAULT_PRIO;
	calc_loops = DEFAULT_CALC_LOOPS;

	rt_init("jhir:t:c:", parse_args, argc, argv);

	if (!period || !prio | !calc_loops) {
		usage();
		exit(1);
	}

	printf("------------------------------------\n");
	printf("Periodic CPU Load Execution Variance\n");
	printf("------------------------------------\n\n");
	printf("Running %d iterations\n", ITERATIONS);
	printf("  priority: %d\n", prio);
	printf("  period: %d ms\n", period/NS_PER_MS);
	printf("  loops: %d\n", calc_loops);

	if (run_jvmsim) {
		printf("jvmsim enabled\n");
		jvmsim_init();	// Start the JVM simulation
	} else {
		printf("jvmsim disabled\n");
	}

	struct periodic_arg parg = { period, ITERATIONS, calc, (void *)calc_loops };

	create_fifo_thread(periodic_thread, (void*)&parg, prio);

	join_threads();

	// FIXME: define pass criteria
	// printf("\nCriteria: latencies < %d us\n", PASS_US);
	// printf("Result: %s\n", ret ? "FAIL" : "PASS");

	return ret;
}
