/*    Filename: tc-2.c
 *      Author: Sripathi Kodi <sripathik@in.ibm.com>
 * Description: Check if clock_gettime is working properly.
 * This test creates NUMSLEEP threads that just sleep and NUMWORK threads
 * that spend time on the CPU. It then reads the thread cpu clocks of all
 * these threads and compares the sum of thread cpu clocks with the process 
 * cpu clock value. The test expects that:
 *   * the cpu clock of every sleeping thread shows close to zero value. 
 *   * sum of cpu clocks of all threads is comparable with the process cpu clock.
 *
 * Compilation: gcc tc-2.c -lpthread -lrt -o tc-2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (C) IBM Corporation, 2007
 *
 * 2007-Apr-04:  Initial version by Sripathi Kodi <sripathik@in.ibm.com>
 */

#include <stdio.h>
#include <pthread.h>
#include <time.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

#define NS_PER_SEC 1000000000
#define THRESHOLD 0.5  /* 500 milliseconds */

#define NUMSLEEP 5
#define NUMWORK 2
struct timespec sleepts[NUMSLEEP];
struct timespec workts[NUMWORK];

/* Just spend some time on the CPU */
void work(void)
{
	unsigned int i=0;
	for (i=0; i<2147483600; i++) {
		if ((i == i+1) || (i == i-1))
			printf("Hey!\n");
	}
}

void *workerthread(void *arg)
{
	int tid = (int)arg;
	struct timespec *ts = &workts[tid];

#ifdef DEBUG
	printf("Worker thread %d working\n", tid);
#endif
	work();

	if ((clock_gettime (CLOCK_THREAD_CPUTIME_ID, ts)) < 0) {
		perror("clock_gettime: CLOCK_THREAD_CPUTIME_ID: ");
		exit(1);
	}

#ifdef DEBUG
	printf("workerthread %d: AFTER WORK: tv_sec = %ld, tv_nsec = %ld\n", tid, ts->tv_sec, ts->tv_nsec);
#endif
	return 0;
}


void *sleeperthread(void *arg)
{
	int tid = (int)arg;
	struct timespec *ts = &sleepts[tid];

#ifdef DEBUG
	printf("Sleeper thread %d sleeping\n", tid);
#endif

	sleep(5);

	if ((clock_gettime (CLOCK_THREAD_CPUTIME_ID, ts)) < 0) {
		perror("clock_gettime: CLOCK_THREAD_CPUTIME_ID: ");
		exit(1);
	}

#ifdef DEBUG
	printf("sleeperthread %d: AFTER SLEEP: tv_sec = %ld, tv_nsec = %ld\n", tid, ts->tv_sec, ts->tv_nsec);
#endif
	return 0;
}

int checkresult(float proctime)
{
	int i, retval = 0;
	float diff, threadstime=0;
	for (i=0; i<NUMSLEEP; i++) {
		/* Sleeping thread should not accumulate more than 1 second of CPU time */
		if (sleepts[i].tv_sec > 0) {
			printf("Sleeper thread %d time is %f, should have been close to zero. FAIL\n", 
				i, sleepts[i].tv_sec + ((float)sleepts[i].tv_nsec/NS_PER_SEC));
			retval = 1;
		}
		threadstime += sleepts[i].tv_sec + ((float)sleepts[i].tv_nsec/NS_PER_SEC);
	}
	if (retval)
		return retval;

	for (i=0; i<NUMWORK; i++) {
		threadstime += workts[i].tv_sec + ((float)workts[i].tv_nsec/NS_PER_SEC);
	}
	printf("\nproctime = %f, threadstime = %f\n", proctime, threadstime);
	diff = proctime - threadstime;
	if (diff < 0) diff = -diff;
	/* Difference between the sum of thread times and process time 
	 * should not be more than THRESHOLD */
	if (diff > THRESHOLD) {
		printf("Diff between threads time and process time is greater than threshold. FAIL.\n");
		retval = 1;
	}
	return retval;
}

int main()
{
	pthread_t sleepthr[NUMSLEEP], workthr[NUMWORK];
	int i, retval = 0;
	struct timespec myts;

	/* Start sleeper threads */
	for (i=0; i<NUMSLEEP; i++) {
		if ((pthread_create (&sleepthr[i], NULL, sleeperthread, (void *)i)) < 0 ) {
			perror("pthread_create: ");
			exit(1);
		}
	}
	printf("\n%d sleeper threads created\n", NUMSLEEP);

	/* Start worker threads */
	for (i=0; i<NUMWORK; i++) {
		if ((pthread_create (&workthr[i], NULL, workerthread, (void *)i)) < 0 ) {
			perror("pthread_create: ");
			exit(1);
		}
	}
	printf("\n%d worker threads created\n", NUMWORK);

	printf("\nPlease wait...\n\n");

	for (i=0; i<NUMSLEEP; i++) {
		pthread_join(sleepthr[i], NULL);
	}
	for (i=0; i<NUMWORK; i++) {
		pthread_join(workthr[i], NULL);
	}

	/* Get the process cpu clock value */
	if ((clock_gettime (CLOCK_PROCESS_CPUTIME_ID, &myts)) < 0) {
		perror("clock_gettime: CLOCK_PROCESS_CPUTIME_ID: ");
		exit(1);
	} 
	retval = checkresult(myts.tv_sec + ((float)myts.tv_nsec/NS_PER_SEC));
	printf("\n Test %s\n\n", retval==0 ? "PASSED" : "FAILED");
	return retval;
}
